(*
 * Hoare Logic,
 * Annotated to Plain conversion
 * 
 * CVS : $Id: tlhlAnnot2Plain.v,v 1.4 2005/03/26 13:11:46 nickie Exp $
*)

Require Export tlhlPlainTyp.
Require Export tlhlAnnotTyp.
Require Export tlhlPlainSem.
Require Export tlhlAnnotSem.
Require Export tlhlPlainLogic.
Require Export tlhlAnnotLogic.

Definition UE : AExpr -> Expr
.
Proof.        
intros ae.
induction ae.
 exact (eint i).
 exact (ebool b).
 exact (evar v).
 exact (eunop u IHae).
 exact (ebinop b IHae1 IHae2).
 exact IHae.
Defined.


Definition UC : AComm -> Comm
.
Proof.
intro ac.
induction ac.
 exact cskip.
 exact (cassign v (UE a)).
 exact (cseq IHac1 IHac2).
 exact (cif (UE a) IHac1 IHac2).
 exact (cwhile (UE a) IHac).
 exact cskip.
Defined.


Theorem typEPreserve :
   forall g  : Env,
   forall ae : AExpr,
   forall t  : Omega,
      typAE g ae t -> typE g (UE ae) t
.
Proof.
intros g ae t d.
induction d.
 simpl in |- *.
   apply typeEint.
 simpl in |- *.
   apply typeEbool.
 simpl in |- *.
   apply typeEvar.
 simpl in |- *.
   apply typeEunop.
   assumption.
 simpl in |- *.
   apply typeEbinop.
  assumption.
  assumption.
 simpl in |- *.
   apply IHd.
Qed.

Theorem typCPreserve :
   forall g  : Env,
   forall ac : AComm,
      typAC g ac -> typC g (UC ac)
.
Proof.
intros g ac d.
induction d.
 simpl in |- *.
   apply typeCskip.
 simpl in |- *.
   apply typeCassign.
   apply typEPreserve.
   assumption.
 simpl in |- *.
   apply typeCseq.
  assumption.
  assumption.
 simpl in |- *.
   apply typeCif.
  apply typEPreserve.
    assumption.
  assumption.
  assumption.
 simpl in |- *.
   apply typeCwhile.
  apply typEPreserve.
    assumption.
  assumption.
 simpl in |- *.
   apply typeCskip.
Qed.


Theorem semEPreserve :
	forall g  : Env,
	forall ae : AExpr,
	forall t  : Omega,
	forall s  : Store g,
   forall v  : semT t,
		semAE g ae t s v -> semE g (UE ae) t s v
.
Proof.
intros g ae t s v se.
induction se.
 simpl in |- *.
   apply semEint.
 simpl in |- *.
   apply semEbool.
 simpl in |- *.
   apply semEvar.
 simpl in |- *.
   apply semEunop.
   assumption.
 simpl in |- *.
   apply semEbinop.
  assumption.
  assumption.
 simpl in |- *.
   assumption.
Qed.


Theorem semCPreserve :
	forall g  : Env,
	forall ac : AComm,
	forall s  : Store g,
	forall s' : Store g,
		semAC g ac s s' -> semC g (UC ac) s s'
.
Proof.
intros g ac s s' sc.
induction sc.
 simpl in |- *.
   apply semCskip.
 simpl in |- *.
   apply semCassign.
   apply semEPreserve.
   assumption.
 simpl in |- *.
   apply semCseq with (s1 := s1).
  assumption.
  assumption.
 simpl in |- *.
   apply semCift.
  apply semEPreserve.
    assumption.
  assumption.
 simpl in |- *.
   apply semCiff.
  apply semEPreserve.
    assumption.
  assumption.
 simpl in |- *.
   apply semCwhilet with (s1 := s1).
  apply semEPreserve.
    assumption.
  assumption.
  assumption.
 simpl in |- *.
   apply semCwhilef.
   apply semEPreserve.
   assumption.
 simpl in |- *.
   apply semCskip.
Qed.


Theorem logicEPreserve :
   forall g  : Env,
   forall ae : AExpr,
   forall t  : Omega,
   forall P  : Pred g,
   forall F  : semT t -> Pred g,
      SpecAE g ae t P F -> SpecE g (UE ae) t P F
.
Proof.
intros g ae t P F d.
induction d.
 simpl in |- *.
   apply axiomEint.
 simpl in |- *.
   apply axiomEbool.
 simpl in |- *.
   apply axiomEvar.
 simpl in |- *.
   apply axiomEunop.
   assumption.
 simpl in |- *.
   apply axiomEbinop with (G := G).
  assumption.
  assumption.
 simpl in |- *.
   apply axiomEconsl with (Q := Q).
  assumption.
  assumption.
Qed. 

Theorem logicCPreserve :
   forall g  : Env,
   forall ac : AComm,
   forall P  : Pred g,
   forall F  : Pred g,
      SpecAC g ac P F -> SpecC g (UC ac) P F
.
Proof.
intros g ac P F d.
induction d.
 simpl in |- *.
   apply axiomCskip.
 simpl in |- *.
   apply axiomCassign.
   apply logicEPreserve.
   assumption.
 simpl in |- *.
   apply axiomCseq with (R := R).
  assumption.
  assumption.
 simpl in |- *.
   apply axiomCif with (F := F).
  apply logicEPreserve.
    assumption.
  assumption.
  assumption.
 simpl in |- *.
   apply axiomCwhile with (F := F).
  apply logicEPreserve.
    assumption.
  assumption.
 simpl in |- *.
   apply axiomCconsl with (Q := Q).
  assumption.
  apply axiomCskip.
Qed.
