(*
 * Hoare Logic,
 * Annotated version --> Typing
 * 
 * CVS : $Id: tlhlAnnotTyp.v,v 1.7 2005/04/11 13:21:48 nickie Exp $
*)

Require Export tlhlAnnotDef.

Inductive typAE (g : Env) : AExpr -> Omega -> Type :=
   typeAEint :
      forall n : Int,
         typAE g (aeint n) uint
 | typeAEbool :
      forall b : Bool,
         typAE g (aebool b) ubool
 | typeAEvar :
      forall x : Var,
         typAE g (aevar x) (g x)
 | typeAEunop :
      forall op : UnOp,
      forall e  : AExpr,
         typAE g e (unopT1 op) ->
            typAE g (aeunop op e) (unopTR op)
 | typeAEbinop :
      forall op : BinOp,
      forall e1 : AExpr,
      forall e2 : AExpr,
         typAE g e1 (binopT1 op) ->
         typAE g e2 (binopT2 op) ->
            typAE g (aebinop op e1 e2) (binopTR op)
 | typeAEweak :
      forall P : Pred g,
      forall Q : Pred g,
      forall p : pred_implies g P Q,
      forall e : AExpr,
      forall t : Omega,
         typAE g e t -> typAE g (aeweak g P Q p e) t
.

Lemma typeALvar:
  forall g : Env,
  forall x : Var,
  forall t : Omega,
    g x = t ->
      typAE g (aevar x) t
.
Proof.
intros g x t H.
rewrite <- H.
apply typeAEvar.
Defined.

Inductive typAC (g : Env) : AComm -> Type :=
   typeACskip : typAC g acskip
 | typeACassign :
      forall x : Var,
      forall e : AExpr,
   typAE g e (g x) ->
      typAC g (acassign x e)
 | typeACseq :
      forall c1 : AComm,
      forall c2 : AComm,
   typAC g c1 ->
   typAC g c2 ->
      typAC g (acseq c1 c2)
 | typeACif :
      forall e  : AExpr,
      forall c1 : AComm,
      forall c2 : AComm,
   typAE g e ubool ->
   typAC g c1 ->
   typAC g c2 ->
      typAC g (acif e c1 c2)
 | typeACwhile :
      forall P : Pred g,
      forall e : AExpr,
      forall c : AComm,
   typAE g e ubool ->
   typAC g c -> typAC g (acwhile g P e c)
 | typeACassert :
   forall P : Pred g,
   forall Q : Pred g,
   forall p : pred_implies g P Q,
      typAC g (acassert g P Q p)
.

Lemma typeALassign:
  forall g : Env,
  forall x : Var,
  forall e : AExpr,
  forall t : Omega,
    typAE g e t ->
    g x = t ->
      typAC g (acassign x e)
.
Proof.
intros g x e t He Hq.
apply typeACassign.
rewrite Hq.
assumption.
Defined.


Hint Constructors typAE typAC : hoare.
Hint Resolve typeALvar typeALassign : hoare.
Hint Extern 2 (typAE ?g (aebinop ?op ?e1 ?e2) ?t) =>
  apply typeAEbinop with (op := op)
.

Ltac typeAC :=
  match goal with
  | |- typAC ?g ?p =>
         unfold p;
         auto 42 with hoare
  | |- _ => idtac
  end
.


Theorem typeAE_uniqueness :
  forall g  : Env,
  forall e  : AExpr,
  forall t1 : Omega,
  forall t2 : Omega,
  forall d1 : typAE g e t1,
  forall d2 : typAE g e t2,
    t1 = t2
.
Proof.
intros g e t1 t2.
induction e.
 intros d1 d2.
   inversion d1; inversion d2; reflexivity.
 intros d1 d2.
   inversion d1; inversion d2; reflexivity.
 intros d1 d2.
   inversion d1; inversion d2; reflexivity.
 destruct u.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
 destruct b.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
  intros d1 d2; inversion d1; inversion d2; reflexivity.
 intros d1 d2.
   inversion d1.
   simpl in X.
   apply IHe.
  rewrite H0.
    assumption.
  inversion d2.
  assumption.
Qed.

(*
Theorem typeAE_deriv_irrel :
  forall g  : Env,
  forall e  : AExpr,
  forall t  : Omega,
  forall d1 : typAE g e t,
  forall d2 : typAE g e t,
    d1 = d2
.
Proof.
Qed.
*)
