(*
 * Hoare Logic,
 * Annotated version --> Weakest Precondition
 * 
 * CVS : $Id: tlhlAnnotWP.v,v 1.7 2005/04/12 01:29:57 nickie Exp $
*)

Require Export tlhlAnnotLogic.
Require Export useful.

Definition wpE :
   forall g : Env,
   forall e : AExpr,
   forall t : Omega,
      typAE g e t ->
         (semT t -> Pred g) -> Pred g
.
Proof.
intros g e t d.
induction d.
 intro F.
   exact (F n).
 intro F.
   exact (F b).
 intro F.
   exact (fun s : Store g => F (s x) s).
 intro F.
   apply IHd.
   exact (fun v : semT (unopT1 op) => F (unopFun op v)).
 intro F.
   apply IHd1.
   intro v1.
   apply IHd2.
   exact (fun v2 : semT (binopT2 op) => F (binopFun op v1 v2)).
 intro F.
   exact P.
Defined.

Definition wpC :
   forall g : Env,
   forall c : AComm,
      typAC g c ->
         Pred g -> Pred g
.
Proof.
intros g c d.
induction d.
 intro Q.
   exact Q.
 intro Q.
   apply wpE with (g := g) (e := e) (t := g x).
  assumption.
  exact (fun (v : semT (g x)) (s : Store g) => Q (updateStore g x v s)).
 intro Q.
   apply IHd1.
   apply IHd2.
   exact Q.
 intro Q.
   apply wpE with (g := g) (e := e) (t := ubool).
  assumption.
  intro b.
    destruct b.
   apply IHd1.
     exact Q.
   apply IHd2.
     exact Q.
 intro Ff.
   set (Ft := IHd P) in |- *.
   set (F := fun b : Bool => if b then Ft else Ff) in |- *.
   set (P' := wpE g e ubool t F).
   exact P.
 intro Q'.
   exact P.
Defined.

Inductive wpEdefined (g : Env):
   forall e : AExpr,
   forall t : Omega,
     typAE g e t -> (semT t -> Pred g) -> Prop
:=
   wpEdef_int :
      forall n : Int,
      forall F : Int -> Pred g,
         wpEdefined g (aeint n) uint (typeAEint g n) F
 | wpEdef_bool :
      forall b : Bool,
      forall F : Bool -> Pred g,
         wpEdefined g (aebool b) ubool (typeAEbool g b) F
 | wpEdef_var :
      forall x : Var,
      forall F : semT (g x) -> Pred g,
         wpEdefined g (aevar x) (g x) (typeAEvar g x) F
 | wpEdef_unop :
      forall op : UnOp,
      forall e  : AExpr,
      forall d  : typAE g e (unopT1 op),
      forall F  : semT (unopTR op) -> Pred g,
      wpEdefined g e (unopT1 op) d
          (fun v : semT (unopT1 op) => F (unopFun op v)) ->
            wpEdefined g (aeunop op e) (unopTR op) (typeAEunop g op e d) F
 | wpEdef_binop :
      forall op : BinOp,
      forall e1 : AExpr,
      forall e2 : AExpr,
      forall d1 : typAE g e1 (binopT1 op),
      forall d2 : typAE g e2 (binopT2 op),
      forall F  : semT (binopTR op) -> Pred g,
    (forall v1 : semT (binopT1 op),
       wpEdefined g e2 (binopT2 op) d2
             (fun v2 : semT (binopT2 op) => F (binopFun op v1 v2))) ->
    wpEdefined g e1 (binopT1 op) d1
          (fun v1 : semT (binopT1 op) =>
                wpE g e2 (binopT2 op) d2
         (fun v2 : semT (binopT2 op) => F (binopFun op v1 v2))) ->
            wpEdefined g (aebinop op e1 e2) (binopTR op)
          (typeAEbinop g op e1 e2 d1 d2) F
 | wpEdef_weak :
      forall P : Pred g,
      forall Q : Pred g,
      forall p : pred_implies g P Q,
      forall e : AExpr,
      forall t : Omega,
      forall d : typAE g e t,
      forall F : semT t -> Pred g,
         wpEdefined g e t d F ->
         pred_equivalent g Q (wpE g e t d F) ->
       wpEdefined g (aeweak g P Q p e) t (typeAEweak g P Q p e t d) F
.    

Lemma wpLdef_var :
   forall g : Env,
   forall x : Var,
   forall t : Omega,
   forall p : g x = t,
   forall F : semT t -> Pred g,
      wpEdefined g (aevar x) t (typeALvar g x t p) F
.
Proof.
intros g x t p.
dependent inversion p.
intro F.
unfold typeALvar in |- *.
replace
 (eq_rect (g x) (fun t0 : Omega => typAE g (aevar x) t0)
    (typeAEvar g x) (g x) (refl_equal (g x))) with
 (typeAEvar g x).
 apply wpEdef_var.
 apply eq_rect_eq with (Q := fun t0 : Omega => typAE g (aevar x) t0).
Qed.

(* If this is removed, for some reason "inversion Hdef." in the following
   proof fails... *)
Derive Inversion coqBug1 with (
   forall g : Env,
   forall e : AExpr,
   forall t : Omega,
      typAE g e t
) Sort Prop.

Theorem wpEisCorrect :
   forall g : Env,
   forall e : AExpr,
   forall t : Omega,
   forall d : typAE g e t,
   forall F : semT t -> Pred g,
      wpEdefined g e t d F ->
         SpecAE g e t (wpE g e t d F) F
.
Proof.
intros g e t d F Hdef.
induction Hdef.
(*int*)
 simpl in |- *.
   apply axiomAEint.
(*bool*)
 simpl in |- *.
   apply axiomAEbool.
(*var*)
 simpl in |- *.
   apply axiomAEvar.
(*unop*)
 simpl in |- *.
   apply axiomAEunop.
   assumption.
(*binop*)
 simpl in |- *.
   apply axiomAEbinop with
       (G := fun v1 : semT (binopT1 op) =>
             wpE g e2 (binopT2 op) d2
               (fun v2 : semT (binopT2 op) => F (binopFun op v1 v2))).
  assumption.
  assumption.
(*weak*)
 simpl in |- *.
   apply axiomAEweak.
   rewrite H.
   assumption.
Qed.

Theorem wpEisExact :
   forall g : Env,
   forall e : AExpr,
   forall t : Omega,
   forall d : typAE g e t,
   forall P : Pred g,
   forall F : semT t -> Pred g,
      SpecAE g e t P F ->
         wpEdefined g e t d F /\
         pred_equivalent g P (wpE g e t d F)
.
Proof.
unfold pred_equivalent in |- *.
intros g e t d.
induction d.
(*int*)
 intros P F Hspec.
   inversion Hspec.
   constructor.
  apply wpEdef_int.
  simpl in |- *.
    replace F with F0.
   reflexivity.
   apply inj_pairT2 with (P := fun x : Omega => semT x -> Pred g) (p := uint).
     assumption.
(*bool*)
 intros P F Hspec.
   inversion Hspec.
   constructor.
  apply wpEdef_bool.
  simpl in |- *.
    replace F with F0.
   reflexivity.
   apply inj_pairT2 with (P := fun x : Omega => semT x -> Pred g) (p := ubool).
     assumption.
(*var*)
 intros P F Hspec.
   inversion Hspec.
   constructor.
  apply wpEdef_var.
  simpl in |- *.
    replace F with F0.
   reflexivity.
   apply inj_pairT2 with (P := fun x : Omega => semT x -> Pred g) (p := g x).
     assumption.
(*unop*)
 intros P F Hspec.
   inversion Hspec.
   destruct (IHd P (fun v : semT (unopT1 op) => F0 (unopFun op v)) H3).
   simpl in |- *.
   assert (F0 = F).
  apply
   inj_pairT2 with (P := fun x : Omega => semT x -> Pred g) (p := unopTR op).
    assumption.
  constructor.
   apply wpEdef_unop.
     rewrite <- H6.
     assumption.
   rewrite <- H6.
     assumption.
(*binop*)
 intros P F Hspec.
   inversion Hspec.
   destruct (IHd1 P G H4).
   simpl in |- *.
   assert (F0 = F).
  apply inj_pairT2 with
      (P := fun x : Omega => semT x -> Pred g)
      (p := binopTR op).
    assumption.
  assert
   (G =
    (fun v1 : semT (binopT1 op) =>
     wpE g e2 (binopT2 op) d2
       (fun v2 : semT (binopT2 op) => F (binopFun op v1 v2)))).
   apply ext.
     intro v1.
     destruct
      (IHd2 (G v1) (fun v2 : semT (binopT2 op) => F0 (binopFun op v1 v2))
         (H5 v1)).
     rewrite <- H8.
     assumption.
   rewrite <- H9.
     constructor.
    apply wpEdef_binop.
     intro v1.
       destruct
        (IHd2 (G v1) (fun v2 : semT (binopT2 op) => F0 (binopFun op v1 v2))
           (H5 v1)).
       rewrite <- H8.
       assumption.
     rewrite <- H9.
       assumption.
    assumption.
(*weak*)
 intros P' F Hspec.
   inversion Hspec.
   simpl in |- *.
   assert (F0 = F).
  apply inj_pairT2 with (P := fun x : Omega => semT x -> Pred g) (p := t).
    assumption.
  destruct (IHd Q0 F0 H4).
    constructor.
   apply wpEdef_weak.
    rewrite <- H6.
      assumption.
    unfold pred_equivalent in |- *.
      rewrite <- H6.
      rewrite <- H8.
      apply inj_pairT2 with (P := Pred).
      symmetry  in |- *.
      assumption.
   rewrite <- H3.
     apply inj_pairT2 with (P := Pred).
     assumption.
Qed.

Inductive wpCdefined (g : Env):
   forall c : AComm,
     typAC g c -> Pred g -> Prop
:=
   wpCdef_skip :
      forall Q : Pred g,
         wpCdefined g acskip (typeACskip g) Q
 | wpCdef_assign :
      forall x  : Var,
      forall e  : AExpr,
      forall de : typAE g e (g x),
      forall Q  : Pred g,
        wpEdefined g e (g x) de 
          (fun (v : semT (g x)) (s : Store g) => Q (updateStore g x v s)) ->
        wpCdefined g (acassign x e) (typeACassign g x e de) Q
 | wpCdef_seq :
      forall c1 : AComm,
      forall c2 : AComm,
      forall d1 : typAC g c1,
      forall d2 : typAC g c2,
      forall Q  : Pred g,
         wpCdefined g c2 d2 Q ->
         wpCdefined g c1 d1 (wpC g c2 d2 Q) ->
         wpCdefined g (acseq c1 c2) (typeACseq g c1 c2 d1 d2) Q
 | wpCdef_if :
      forall e  : AExpr,
      forall c1 : AComm,
      forall c2 : AComm,
      forall de : typAE g e ubool,
      forall d1 : typAC g c1,
      forall d2 : typAC g c2,
      forall Q  : Pred g,
         wpCdefined g c1 d1 Q -> wpCdefined g c2 d2 Q ->
         wpEdefined g e ubool de (fun b : semT ubool => if b then
                 wpC g c1 d1 Q
            else
                 wpC g c2 d2 Q) ->
         wpCdefined g (acif e c1 c2) (typeACif g e c1 c2 de d1 d2) Q
 | wpCdef_while :
      forall e  : AExpr,
      forall c  : AComm,
      forall P  : Pred g,
      forall dc : typAC g c, 
      forall de : typAE g e ubool,
      forall Q  : Pred g,
         wpCdefined g c dc P ->
         wpEdefined g e ubool de (fun b : semT ubool => 
            if b then wpC g c dc P else Q) ->
         pred_equivalent g P
           (wpE g e ubool de (fun b : semT ubool =>
              if b then wpC g c dc P else Q)) ->
         wpCdefined g (acwhile g P e c) (typeACwhile g P e c de dc) Q
 | wpCdef_assert :
      forall P  : Pred g,
      forall Q  : Pred g,
      forall p  : pred_implies g P Q,
      forall Q' : Pred g,
         pred_equivalent g Q Q' ->
         wpCdefined g (acassert g P Q p) (typeACassert g P Q p) Q'
. 

Theorem wpCisCorrect :
   forall g : Env,
   forall c : AComm,
   forall d : typAC g c,
   forall Q : Pred g,
      wpCdefined g c d Q ->
         SpecAC g c (wpC g c d Q) Q
.
Proof.
intros g c d Q Hdef.
induction Hdef.
(*skip*)
 simpl in |- *.
   apply axiomACskip.
(*assign*)
 simpl in |- *.
   apply axiomACassign.
   apply wpEisCorrect.
   assumption.
(*seq*)
 simpl in |- *.
   apply axiomACseq with (R := wpC g c2 d2 Q).
  assumption.
  assumption.
(*if*)
 simpl in |- *.
   apply axiomACif with
       (F := fun b : Bool => if b then wpC g c1 d1 Q else wpC g c2 d2 Q).
  apply wpEisCorrect.
    assumption.
  assumption.
  assumption.
(*while*)
 simpl in |- *.
   apply axiomACwhile with
       (F := fun b : semT ubool => if b then wpC g c dc P else Q).
  set (F := fun b : semT ubool => if b then wpC g c dc P else Q) in |- *.
    rewrite H0.
    fold F in |- *.
    apply wpEisCorrect.
    unfold F in |- *.
    assumption.
  assumption.
(*assert*)
 simpl in |- *.
   replace Q' with Q.
   apply axiomACassert.
Qed.

Theorem wpCisExact :
   forall g : Env,
   forall c : AComm,
   forall d : typAC g c,
   forall P : Pred g,
   forall Q : Pred g,
      SpecAC g c P Q ->
        wpCdefined g c d Q /\ pred_equivalent g P (wpC g c d Q)
.
Proof.
unfold pred_equivalent.
intros g c d.
induction d.
(*skip*)
 intros P Q Hspec.
   inversion Hspec.
   constructor.
  apply wpCdef_skip.
  simpl in |- *.
    reflexivity.
(*assign*)
 intros P Q Hspec.
   inversion Hspec.
   destruct
    (wpEisExact g e (g x) t P
       (fun (v : semT (g x)) (s : Store g) => Q (updateStore g x v s))).
  assumption.
  constructor.
   apply wpCdef_assign.
     assumption.
   simpl in |- *.
     assumption.
(*seq*)
 intros P Q Hspec.
   inversion Hspec.
   simpl in |- *.
   destruct (IHd1 P R H1).
   destruct (IHd2 R Q H4).
   constructor.
  apply wpCdef_seq.
   assumption.
   rewrite <- H8.
     assumption.
  rewrite <- H8.
    assumption.
(*if*)
 intros P Q Hspec.
   inversion Hspec.
   simpl in |- *.
   destruct (wpEisExact g e ubool t P F).
  assumption.
  destruct (IHd1 (F true) Q H5).
    destruct (IHd2 (F false) Q H6).
    assert
     (F =
      (fun b : semT ubool => if b then wpC g c1 d1 Q else wpC g c2 d2 Q)).
   apply ext.
     intro b.
     induction b.
    assumption.
    assumption.
   constructor.
    apply wpCdef_if.
     assumption.
     assumption.
     rewrite <- H13.
       assumption.
    rewrite H8.
      rewrite H13.
      reflexivity.
(*while*)
 intros P' Q Hspec.
   inversion Hspec.
   simpl in |- *.
   assert (P' = P).
  rewrite <- H1.
    apply inj_pairT2 with (P := Pred).
    assumption.
  constructor.
   destruct (wpEisExact g e ubool t P F).
    rewrite <- H6.
      assumption.
    destruct (IHd (F true) P' H5).
      assert (F = (fun b : semT ubool => if b then wpC g c d P' else Q)).
     apply ext.
       intro b.
       induction b.
      assumption.
      assumption.
     apply wpCdef_while.
      rewrite <- H6.
        assumption.
      rewrite <- H6.
        rewrite H3.
        rewrite <- H11.
        assumption.
      unfold pred_equivalent in |- *.
        rewrite H6 in H11.
        rewrite H3.
        rewrite <- H11.
        assumption.
   assumption.
(*assert*)
 intros P' Q' Hspec.
   inversion Hspec.
   simpl in |- *.
   constructor.
  apply wpCdef_assert.
    unfold pred_equivalent in |- *.
    rewrite <- H3.
    symmetry  in |- *.
    apply inj_pairT2 with (P := Pred).
    assumption.
  rewrite <- H2.
    apply inj_pairT2 with (P := Pred).
    assumption.
Qed.

Hint Constructors wpEdefined wpCdefined : hoare.
Hint Resolve wpLdef_var : hoare.

Hint Extern 2 (wpEdefined ?g (aebinop ?op ?e1 ?e2) ?t ?d ?F) =>
  apply wpEdef_binop with (op := op)
.

Ltac specAC d :=
  match goal with
  | |- SpecAC ?g ?p ?P ?Q =>
         apply wpCisCorrect with (d := d) (Q := Q);
         unfold p, d;
         auto 42 with hoare
  | |- _ => idtac
  end
.
