(*
 * Hoare Logic,
 * Definitions
 * 
 * CVS : $Id: tlhlDefinitions.v,v 1.4 2005/04/12 01:29:57 nickie Exp $
*)

Require Export Decidable.
Require Export Arith.
Require Export ZArith.
Require Export Eqdep.
Require Export Bool.
Open Scope nat_scope.
Open Scope Z_scope.

Definition Int  : Set := Z.
Definition Bool : Set := bool.


Definition Var : Set := nat.

Inductive UnOp : Set :=
    opNeg : UnOp
  | opNot : UnOp
.

Inductive BinOp : Set :=
    opPlus  : BinOp
  | opMinus : BinOp
  | opTimes : BinOp
  | opDiv   : BinOp
  | opMod   : BinOp
  | opEq    : BinOp
  | opNe    : BinOp
  | opLt    : BinOp
  | opLe    : BinOp
  | opGt    : BinOp
  | opGe    : BinOp
  | opAnd   : BinOp
  | opOr    : BinOp
.

Inductive Omega : Set :=
    uint   : Omega
  | ubool  : Omega
.

Definition Env : Set :=
   Var -> Omega
.

Definition semT (t : Omega) : Set :=
   match t with
      uint  => Int
    | ubool => Bool
   end
.

Definition Store : Env -> Set :=
   fun g : Env => forall x : Var, semT (g x)
.

Definition updateStore :
   forall g : Env, forall x : Var,
      semT (g x) -> Store g -> Store g
:=
   fun g : Env =>
   fun x : Var =>
   fun v : semT (g x) =>
   fun s : Store g =>
   fun y : Var =>
      match eq_nat_dec x y with
         left H  => eq_rec x (fun y : Var => semT (g y)) v y H
       | right _ => s y
      end
.

Definition Pred : Env -> Type :=
   fun g : Env => Store g -> Prop (*Set*)
.

Definition SPECE :=
   fun g : Env =>
   fun t : Omega =>
      Pred g -> (semT t -> Pred g) -> Prop (*Set*) (*Type*)
.

Definition SPECC :=
   fun g : Env =>
      Pred g -> Pred g -> Prop (*Set*) (*Type*)
.

Definition pred_implies :
   forall g : Env, Pred g -> Pred g -> Prop
:=
   fun g : Env =>
   fun P : Pred g =>
   fun Q : Pred g =>
      forall s : Store g, P s -> Q s
.

(* We take $\alpha\beta\iota$-equality here *)
Definition pred_equivalent :
   forall g : Env, Pred g -> Pred g -> Prop
:=
   fun g : Env =>
   fun P : Pred g =>
   fun Q : Pred g =>
      P = Q
.

Definition unopT1 : UnOp -> Omega :=
   fun op : UnOp =>
   match op with
      opNeg => uint
    | opNot => ubool
   end
.
     
Definition unopTR : UnOp -> Omega :=
   fun op : UnOp =>
   match op with
      opNeg => uint
    | opNot => ubool
   end
.
     
Definition unopFun :
   forall op : UnOp,
      semT (unopT1 op) -> semT (unopTR op)
:=
   fun op : UnOp =>
   match op
   return semT (unopT1 op) -> semT (unopTR op)
   with
      opNeg => Zopp
    | opNot => negb
   end
.

Definition binopT1 : BinOp -> Omega :=
   fun op : BinOp =>
   match op with
      opPlus  => uint
    | opMinus => uint
    | opTimes => uint
    | opDiv   => uint
    | opMod   => uint
    | opEq    => uint
    | opNe    => uint
    | opLt    => uint
    | opLe    => uint
    | opGt    => uint
    | opGe    => uint
    | opAnd   => ubool
    | opOr    => ubool
   end
.
     
Definition binopT2 : BinOp -> Omega :=
   fun op : BinOp =>
   match op with
      opPlus  => uint
    | opMinus => uint
    | opTimes => uint
    | opDiv   => uint
    | opMod   => uint
    | opEq    => uint
    | opNe    => uint
    | opLt    => uint
    | opLe    => uint
    | opGt    => uint
    | opGe    => uint
    | opAnd   => ubool
    | opOr    => ubool
   end
.
     
Definition binopTR : BinOp -> Omega :=
   fun op : BinOp =>
   match op with
      opPlus  => uint
    | opMinus => uint
    | opTimes => uint
    | opDiv   => uint
    | opMod   => uint
    | opEq    => ubool
    | opNe    => ubool
    | opLt    => ubool
    | opLe    => ubool
    | opGt    => ubool
    | opGe    => ubool
    | opAnd   => ubool
    | opOr    => ubool
   end
.
     
Definition binopFun :
   forall op : BinOp,
      semT (binopT1 op) -> semT (binopT2 op) -> semT (binopTR op)
:=
   fun op : BinOp =>
   match op
   return semT (binopT1 op) -> semT (binopT2 op) -> semT (binopTR op)
   with
      opPlus  => Zplus
    | opMinus => Zminus
    | opTimes => Zmult
    | opDiv   => Zdiv
    | opMod   => Zmod
    | opEq    => Zeq_bool
    | opNe    => Zneq_bool
    | opLt    => Zlt_bool
    | opLe    => Zle_bool
    | opGt    => Zgt_bool
    | opGe    => Zge_bool
    | opAnd   => andb
    | opOr    => orb
   end
.

Hint Unfold pred_implies.
Hint Unfold pred_equivalent.
