(*
 * Hoare Logic,
 * Plain version --> Logic
 * 
 * CVS : $Id: tlhlPlainLogic.v,v 1.3 2005/03/26 13:11:46 nickie Exp $
*)

Require Export tlhlPlainDef.
Require Export tlhlPlainTyp.
Require Export tlhlPlainSem.

Inductive SpecE (g : Env) :
   forall e : Expr, forall t : Omega, SPECE g t
:=
    axiomEint :
       forall n : Int,
       forall F : Int -> Pred g,
         SpecE g (eint n) uint (F n) F
  | axiomEbool :
       forall b : Bool,
       forall F : Bool -> Pred g,
         SpecE g (ebool b) ubool (F b) F
  | axiomEvar :
       forall x : Var,
       forall F : semT (g x) -> Pred g,
         SpecE g (evar x) (g x) (fun s : Store g => F (s x) s) F
  | axiomEunop :
       forall op : UnOp,
       forall e  : Expr,
       forall P  : Pred g,
       forall F  : semT (unopTR op) -> Pred g,
       SpecE g e (unopT1 op) P
           (fun v : semT (unopT1 op) => F (unopFun op v)) ->
         SpecE g (eunop op e) (unopTR op) P F
  | axiomEbinop :
      forall op : BinOp,
      forall e1 : Expr,
      forall e2 : Expr,
      forall P  : Pred g,
      forall F  : semT (binopTR op) -> Pred g,
      forall G  : semT (binopT1 op) -> Pred g,
      SpecE g e1 (binopT1 op) P G ->
      (forall v1 : semT (binopT1 op),
              SpecE g e2 (binopT2 op) (G v1)
                (fun v2 : semT (binopT2 op) => F (binopFun op v1 v2))) ->
        SpecE g (ebinop op e1 e2) (binopTR op) P F
  | axiomEconsl :
       forall e : Expr,
       forall t : Omega,
       forall P : Pred g,
       forall Q : Pred g,       
       forall F : semT t -> Pred g,
       forall p : pred_implies g P Q,
           SpecE g e t Q F -> SpecE g e t P F
.

Inductive SpecC (g : Env) :
   forall c : Comm, SPECC g
:=
   axiomCskip :
     forall P : Pred g,
       SpecC g cskip P P
 | axiomCassign :
     forall x  : Var,
     forall e  : Expr,
     forall P  : Pred g,
     forall Q  : Pred g,
     SpecE g e (g x) P
         (fun v : semT (g x) => fun s : Store g => Q (updateStore g x v s)) ->
       SpecC g (cassign x e) P Q
 | axiomCseq :
     forall c1 : Comm,
     forall c2 : Comm,
     forall P  : Pred g,
     forall Q  : Pred g,
     forall R  : Pred g,
     SpecC g c1 P R -> SpecC g c2 R Q ->
       SpecC g (cseq c1 c2) P Q
 | axiomCif :
     forall e  : Expr,
     forall c1 : Comm,
     forall c2 : Comm,
     forall P  : Pred g,
     forall Q  : Pred g,
     forall F  : semT ubool -> Pred g,
     SpecE g e ubool P F -> SpecC g c1 (F true) Q -> SpecC g c2 (F false) Q -> 
        SpecC g (cif e c1 c2) P Q
 | axiomCwhile :
     forall P  : Pred g,
     forall e  : Expr,
     forall c  : Comm,
     forall F  : semT ubool -> Pred g,
     SpecE g e ubool P F -> SpecC g c (F true) P ->
           SpecC g (cwhile e c) P (F false)
 | axiomCconsl :
      forall c : Comm,
      forall P : Pred g,
      forall Q : Pred g,
      forall R : Pred g,
      forall p : pred_implies g P Q,
         SpecC g c Q R -> SpecC g c P R
 | axiomCconsr :
      forall c : Comm,
      forall P : Pred g,
      forall Q : Pred g,
      forall R : Pred g,
      forall p : pred_implies g R P,
         SpecC g c Q R -> SpecC g c Q P
.
