(*
 * Hoare Logic,
 * Plain version --> Semantics
 * 
 * CVS : $Id: tlhlPlainSem.v,v 1.5 2005/03/26 13:11:46 nickie Exp $
*)

Require Export tlhlPlainDef.

Inductive semE (g : Env) :
   forall e : Expr,
   forall t : Omega,
      Store g -> semT t -> Type
:=
   semEint :
      forall n : Int,
      forall s : Store g,
         semE g (eint n) uint s n
 | semEbool :
      forall b : Bool,
      forall s : Store g,
         semE g (ebool b) ubool s b
 | semEvar :
      forall x : Var,
      forall s : Store g,
         semE g (evar x) (g x) s (s x)
 | semEunop :
      forall op : UnOp,
      forall e  : Expr,
      forall s  : Store g,
      forall v  : semT (unopT1 op),
      semE g e (unopT1 op) s v ->
         semE g (eunop op e) (unopTR op) s (unopFun op v)
 | semEbinop :
      forall op : BinOp,
      forall e1 : Expr,
      forall e2 : Expr,
      forall s  : Store g,
      forall v1 : semT (binopT1 op),
      forall v2 : semT (binopT2 op),
      semE g e1 (binopT1 op) s v1 ->
      semE g e2 (binopT2 op) s v2 ->
         semE g (ebinop op e1 e2) (binopTR op) s (binopFun op v1 v2)
.

Inductive semC (g : Env) :
   forall c : Comm,
      Store g -> Store g -> Type
:=
   semCskip :
      forall s : Store g,
         semC g cskip s s
 | semCassign :
      forall x  : Var,
      forall e  : Expr,
      forall s : Store g,
      forall v : semT (g x),
      semE g e (g x) s v ->
         semC g (cassign x e)
              s (updateStore g x v s)
 | semCseq :
      forall c1 : Comm,
      forall c2 : Comm,
      forall s  : Store g,
      forall s1 : Store g,
      forall s2 : Store g,
      semC g c1 s s1 ->
      semC g c2 s1 s2 ->
         semC g (cseq c1 c2) s s2
 | semCift :
      forall e  : Expr,
      forall c1 : Comm,
      forall c2 : Comm,
      forall s  : Store g,
      forall s1 : Store g,
      semE g e ubool s true ->
      semC g c1 s s1 ->
         semC g (cif e c1 c2) s s1
 | semCiff :
      forall e  : Expr,
      forall c1 : Comm,
      forall c2 : Comm,
      forall s  : Store g,
      forall s2 : Store g,
      semE g e ubool s false ->
      semC g c2 s s2 ->
         semC g (cif e c1 c2) s s2
 | semCwhilet :
      forall e  : Expr,
      forall c  : Comm,
      forall s  : Store g,
      forall s1 : Store g,
      forall s2 : Store g,
      semE g e ubool s true ->
      semC g c s s1 ->
      semC g (cwhile e c) s1 s2 ->
         semC g (cwhile e c) s s2
 | semCwhilef :
      forall e  : Expr,
      forall c  : Comm,
      forall s  : Store g,
      semE g e ubool s false ->
         semC g (cwhile e c) s s
.
