(*
 * Hoare Logic,
 * Plain version --> Typing
 * 
 * CVS : $Id: tlhlPlainTyp.v,v 1.5 2005/03/26 13:11:46 nickie Exp $
*)

Require Export tlhlPlainDef.

Inductive typE (g : Env) : Expr -> Omega -> Type :=
   typeEint :
      forall n : Int,
         typE g (eint n) uint
 | typeEbool :
      forall b : Bool,
         typE g (ebool b) ubool
 | typeEvar :
      forall x : Var,
         typE g (evar x) (g x)
 | typeEunop :
      forall op : UnOp,
      forall e  : Expr,
         typE g e (unopT1 op) ->
            typE g (eunop op e) (unopTR op)
 | typeEbinop :
      forall op : BinOp,
      forall e1 : Expr,
      forall e2 : Expr,
         typE g e1 (binopT1 op) ->
         typE g e2 (binopT2 op) ->
            typE g (ebinop op e1 e2) (binopTR op)
.

Lemma typeLvar:
  forall g : Env,
  forall x : Var,
  forall t : Omega,
    g x = t ->
      typE g (evar x) t
.
Proof.
intros g x t H.
rewrite <- H.
apply typeEvar.
Defined.

Inductive typC (g : Env) : Comm -> Type :=
   typeCskip : typC g cskip
 | typeCassign :
      forall x : Var,
      forall e : Expr,
      typE g e (g x) ->
      typC g (cassign x e)
 | typeCseq :
      forall c1 : Comm,
      forall c2 : Comm,
      typC g c1 ->
      typC g c2 ->
      typC g (cseq c1 c2)
 | typeCif :
      forall e  : Expr,
      forall c1 : Comm,
      forall c2 : Comm,
      typE g e ubool ->
      typC g c1 ->
      typC g c2 ->
      typC g (cif e c1 c2)
 | typeCwhile :
      forall e : Expr,
      forall c : Comm,
      typE g e ubool ->
      typC g c -> typC g (cwhile e c)
.

Lemma typeLassign:
  forall g : Env,
  forall x : Var,
  forall e : Expr,
  forall t : Omega,
    typE g e t ->
    g x = t ->
      typC g (cassign x e)
.
Proof.
intros g x e t He Hq.
apply typeCassign.
rewrite Hq.
assumption.
Defined.

Hint Constructors typE typC : hoare.
Hint Resolve typeLvar typeLassign : hoare.
