Require Export ZArith.
Require Export Eqdep.
Require Export Setoid.

Open Scope Z_scope.


Section Extensionality.

Definition extensionality :
   forall A : Type,
   forall B : A -> Type,
     Prop
:=
   fun A : Type =>
   fun B : A -> Type =>
     forall F : (forall x : A, B x),
     forall G : (forall x : A, B x),
       (forall x : A, F x = G x) ->
         F = G
.

End Extensionality.


Section DependentEquality.

Theorem eq_dep_2 :
   forall A  : Type,
   forall P  : A -> Type,
   forall Q  : A -> Type,
   forall x1 : A,
   forall x2 : A,
   forall y1 : P x1,
   forall y2 : P x2,
   forall z1 : Q x1,
   forall z2 : Q x2,
      existT P x1 y1 = existT P x2 y2 ->
      existT Q x1 z1 = existT Q x2 z2 ->
         forall H : forall x : A, P x -> Q x -> Prop,
            H x1 y1 z1 -> H x2 y2 z2
.
Proof.
intros A P Q x1 x2 y1 y2 z1 z2 H1 H2 H.
set (H' := fun (a b : A) (p : P a) (q : Q b) =>
           forall e : a = b, H b (eq_rect a P p b e) q).
assert (forall (x : A) (p : P x) (q : Q x), H x p q <-> H' x x p q).
 intros x p q.
   unfold H' in |- *.
   split.
  intros h e.
    replace (eq_rect x P p x e) with p.
   assumption.
   apply eq_rect_eq.
  intro Z.
    replace p with (eq_rect x P p x (refl_equal x)).
   apply Z.
   symmetry  in |- *.
     apply eq_rect_eq.
 setoid_rewrite (H0 x1 y1 z1).
   setoid_rewrite (H0 x2 y2 z2).
   intro h.
   change y2 with (projT2 (existT P x2 y2)) in |- *.
   change x2 at 1 with (projT1 (existT P x2 y2)) in |- *.
   rewrite <- H1.
   simpl in |- *.
   change z2 with (projT2 (existT Q x2 z2)) in |- *.
   change x2 at 1 with (projT1 (existT Q x2 z2)) in |- *.
   rewrite <- H2.
   simpl in |- *.
   assumption.
Qed.

Theorem f_equal_dep :
   forall A : Type,
   forall B : A -> Type,
   forall f : forall a : A, B a,
   forall x : A,
   forall y : A,
   forall p : x = y,
      eq_rect x B (f x) y p = f y
.
Proof.
intros A B f x y p.
dependent inversion p.
symmetry  in |- *.
apply eq_rect_eq.
Qed.

Theorem eq_ind_dep :
   forall A : Type,
   forall B : A -> Type,
   forall H : forall a : A, B a -> Prop,
   forall x : A,
   forall y : A,
   forall z : B x,
   forall p : x = y,
     H x z ->
       H y (eq_rect x B z y p)
.
Proof.
intros A B H x y z p h.
dependent inversion p.
replace (eq_rect x B z x (refl_equal x)) with z.
 assumption.
 apply eq_rect_eq.
Qed.

End DependentEquality.


Section Mathematics.

Lemma Zmod_by_1 :
   forall x : Z,
      x mod 1 = 0
.
Proof.
intro x.
assert (0 <= x mod 1 < 1).
 apply Z_mod_lt.
   trivial with zarith.
 destruct H.
   apply Zle_antisym.
  apply Zlt_succ_le.
    assumption.
  assumption.
Qed.

Lemma Zdiv_by_1 :
   forall x : Z,
     x / 1 = x
.
Proof.
intro x.
assert (1 * (x / 1) + x mod 1 = x).
 symmetry  in |- *.
   apply Z_div_mod_eq.
   trivial with zarith.
 transitivity (1 * (x / 1) + x mod 1).
  transitivity (1 * (x / 1)).
   symmetry  in |- *.
     apply Zmult_1_l.
   transitivity (1 * (x / 1) + 0).
    symmetry  in |- *.
      apply Zplus_0_r.
    assert (x mod 1 = 0).
     apply Zmod_by_1.
     rewrite H0.
       reflexivity.
  assumption.
Qed.

Lemma Zdiv_lt_1 :
   forall x : Z,
   forall y : Z,
      y > 0 ->
      x < y ->
         x / y < 1
.
Proof.
intros x y Hy Hxy.
apply Znot_ge_lt.
intro Ha.
apply Zle_not_lt with (m := x) (n := y).
 apply Zge_le.
   replace y with (1 * y).
  apply Zge_trans with (m := x / y * y).
   apply Zle_ge.
     replace (x / y * y) with (y * (x / y)).
    apply Z_mult_div_ge.
      assumption.
    apply Zmult_comm.
   apply Zmult_ge_compat_r.
    assumption.
    apply Zle_ge.
      apply Zlt_le_weak.
      apply Zgt_lt.
      assumption.
  apply Zmult_1_l.
 assumption.
Qed.

Lemma Zdiv_zero :
   forall x : Z,
   forall y : Z,
      0 <= x < y ->
         x / y = 0
.
Proof.
intros x y H.
destruct H as (H1, H2).
apply Zle_antisym.
 apply Zlt_succ_le.
   apply Zdiv_lt_1.
  apply Zlt_gt.
    apply Zle_lt_trans with (m := x).
   assumption.
   assumption.
  assumption.
 apply Zge_le.
   apply Z_div_ge0.
  apply Zlt_gt.
    apply Zle_lt_trans with (m := x).
   assumption.
   assumption.
  apply Zle_ge.
    assumption.
Qed.

Lemma Zmod_same :
   forall x : Z,
   forall y : Z,
      0 <= x < y ->
         x mod y = x
.
Proof.
intros x y H.
transitivity (y * (x / y) + x mod y).
 replace (x / y) with 0.
  auto with zarith.
  symmetry  in |- *.
    apply Zdiv_zero.
    assumption.
 symmetry  in |- *.
   apply Z_div_mod_eq.
   apply Zlt_gt.
   destruct H as (H1, H2).
   apply Zle_lt_trans with (m := x).
  assumption.
  assumption.
Qed.

Lemma Zdiv_unique :
   forall x : Z,
   forall y : Z,
   forall a : Z,
   forall b : Z,
      y > 0 ->
      0 <= b < y ->
      x = y * a + b ->
         x / y = a
.
Proof.
intros x y a b Hy Hb Hx.
apply Zminus_eq.
transitivity (y * (x / y - a) / y).
 replace (y * (x / y - a)) with ((x / y - a) * y).
  symmetry  in |- *.
    apply Z_div_mult.
    assumption.
  apply Zmult_comm.
 replace (y * (x / y - a)) with (b - x mod y).
  rewrite Hx.
    replace ((y * a + b) mod y) with (b mod y).
   replace (b - b mod y) with 0.
    trivial.
    symmetry  in |- *.
      apply Zeq_minus.
      symmetry  in |- *.
      apply Zmod_same.
      assumption.
   symmetry  in |- *.
     replace (y * a + b) with (b + a * y).
    apply Z_mod_plus.
      assumption.
    replace (a * y) with (y * a).
     apply Zplus_comm.
     apply Zmult_comm.
  transitivity (y * (x / y) - y * a).
   apply Zplus_minus_eq.
     transitivity (y * a + b - x mod y).
    apply Zplus_minus_eq.
      rewrite <- Hx.
      transitivity (y * (x / y) + x mod y).
     apply Z_div_mod_eq.
       assumption.
     apply Zplus_comm.
    auto with zarith.
   symmetry  in |- *.
     apply Zmult_minus_distr_l.
Qed.

Lemma Zmult_div_assoc :
   forall x : Z,
   forall y : Z,
   forall z : Z,
      y > 0 ->
      z > 0 ->
         x / (y * z) = x / y / z
.
Proof.
intros x y z Hy Hz.
destruct (Z_mod_lt (x / y) z).
 assumption.
 destruct (Z_mod_lt x y).
  assumption.
  apply Zdiv_unique with (b := y * ((x / y) mod z) + x mod y).
   apply Zmult_gt_0_compat.
    assumption.
    assumption.
   constructor.
    apply Zplus_le_0_compat.
     apply Zmult_le_0_compat.
      apply Zlt_le_weak.
        apply Zgt_lt.
        assumption.
      assumption.
     assumption.
    apply Zgt_lt.
      apply Zlt_succ_gt.
      assert (y * ((x / y) mod z) + x mod y <= y * (z - 1) + (y - 1)).
     apply Zplus_le_compat.
      apply Zmult_le_compat_l.
       auto with zarith.
       auto with zarith.
      auto with zarith.
     apply Zle_trans with (m := y * (z - 1) + y).
      change (y * ((x / y) mod z) + x mod y + 1 <= y * (z - 1) + y) in |- *.
        replace (y * (z - 1) + y) with (y * (z - 1) + (y - 1) + 1).
       apply Zplus_le_compat_r.
         assumption.
       transitivity (y * (z - 1) + (y - 1 + 1)).
        symmetry  in |- *.
          apply Zplus_assoc.
        auto with zarith.
      replace (y * (z - 1)) with (y * z - y * 1).
       auto with zarith.
       symmetry  in |- *.
         apply Zmult_minus_distr_l.
   replace (y * z * (x / y / z) + (y * ((x / y) mod z) + x mod y)) with
    (y * (z * (x / y / z) + (x / y) mod z) + x mod y).
    replace (z * (x / y / z) + (x / y) mod z) with (x / y).
     apply Z_div_mod_eq.
       assumption.
     apply Z_div_mod_eq.
       assumption.
    replace (y * z * (x / y / z) + (y * ((x / y) mod z) + x mod y)) with
     (y * z * (x / y / z) + y * ((x / y) mod z) + x mod y).
     replace (y * (z * (x / y / z) + (x / y) mod z)) with
      (y * z * (x / y / z) + y * ((x / y) mod z)).
      reflexivity.
      replace (y * z * (x / y / z)) with (y * (z * (x / y / z))).
       symmetry  in |- *.
         apply Zmult_plus_distr_r.
       apply Zmult_assoc.
     symmetry  in |- *.
       apply Zplus_assoc.
Qed.

Lemma Zpow_gt_zero :
   forall x : Z,
   forall y : Z,
      x > 0 -> y >= 0 -> x^y > 0
.
Proof.
intros x y Hx Hy.
destruct y.
 compute in |- *; reflexivity.
 unfold Zpower in |- *.
   induction p.
  replace (xI p) with (p + p + 1)%positive.
   replace (Zpower_pos x (p + p + 1)) with
    (Zpower_pos x p * Zpower_pos x (p + 1)).
    replace (Zpower_pos x (p + 1)) with (Zpower_pos x p * Zpower_pos x 1).
     apply Zmult_gt_0_compat.
      apply IHp.
        trivial.
      apply Zmult_gt_0_compat.
       apply IHp.
         trivial.
       replace (Zpower_pos x 1) with x.
        assumption.
        unfold Zpower_pos in |- *.
          unfold iter_pos in |- *.
          symmetry  in |- *.
          apply Zmult_1_r.
     symmetry  in |- *.
       apply Zpower_pos_is_exp.
    symmetry  in |- *.
      replace (p + p + 1)%positive with (p + (p + 1))%positive.
     apply Zpower_pos_is_exp.
     apply Pplus_assoc.
   replace (xI p) with (xO p + 1)%positive.
    replace (p + p)%positive with (xO p).
     reflexivity.
     symmetry  in |- *.
       apply Pplus_diag.
    replace (xO p + 1)%positive with (Psucc (xO p)).
     symmetry  in |- *.
       apply xI_succ_xO.
     apply Pplus_one_succ_r.
  replace (xO p) with (p + p)%positive.
   replace (Zpower_pos x (p + p)) with (Zpower_pos x p * Zpower_pos x p).
    apply Zmult_gt_0_compat.
     apply IHp.
       trivial.
     apply IHp.
       trivial.
    symmetry  in |- *.
      apply Zpower_pos_is_exp.
   apply Pplus_diag.
  unfold Zpower_pos in |- *.
    unfold iter_pos in |- *.
    replace (x * 1) with x.
   assumption.
   symmetry  in |- *.
     apply Zmult_1_r.
 contradiction Hy.
   compute in |- *.
   reflexivity.
Qed.

Lemma Zdiv_pow2_1 :
   forall i : Z,
   forall x : Z,
      i >= 0 -> (
         x / (2^i) = 1  ->  2^i <= x < 2^(i+1)
      )
.
Proof.
intros i x Hi.
intro H.
constructor.
 apply Znot_gt_le.
   intro Ha.
   absurd (x / 2 ^ i = 1).
  apply Zlt_not_eq.
    apply Zdiv_lt_1.
   apply Zpow_gt_zero.
    auto with zarith.
    assumption.
   apply Zgt_lt.
     assumption.
  assumption.
 apply Znot_ge_lt.
   intro Ha.
   absurd (1 = x / 2 ^ i).
  apply Zlt_not_eq.
    apply Zgt_lt.
    apply Zlt_succ_gt.
    apply Zge_le.
    replace (Zsucc 1) with (2 ^ (i + 1) / 2 ^ i).
   apply Z_div_ge.
    apply Zpow_gt_zero.
     trivial with zarith.
     assumption.
    assumption.
   replace (i + 1) with (1 + i).
    replace (2 ^ (1 + i)) with (2 ^ 1 * 2 ^ i).
     apply Z_div_mult.
       apply Zpow_gt_zero.
      auto with zarith.
      assumption.
     symmetry  in |- *.
       apply Zpower_exp.
      auto with zarith.
      assumption.
    apply Zplus_comm.
  symmetry  in |- *.
    assumption.
Qed.

Theorem iter_nat_better_invariant :
 forall (n:nat) (A:Set) (f:A -> A) (Inv:nat -> A -> Prop),
   (forall (n:nat) (x:A), Inv n x -> Inv (S n) (f x)) ->
   forall x:A, Inv O x -> Inv n (iter_nat n A f x).
Proof.
simple induction n; intros;
 [ trivial with arith
 | simpl in |- *; apply H0 with (x:= iter_nat n0 A f x); apply H;
    trivial with arith ].
Qed.

Theorem iter_pos_better_invariant :
 forall (p:positive) (A:Set) (f:A -> A) (Inv:nat -> A -> Prop),
   (forall (n:nat) (x:A), Inv n x -> Inv (S n) (f x)) ->
   forall x:A, Inv O x -> Inv (nat_of_P p) (iter_pos p A f x).
Proof.
intros.
rewrite iter_nat_of_P.
apply iter_nat_better_invariant.
 assumption.
 assumption.
Qed.

Definition Zfact : Z -> Z :=
  fun n : Z =>
    fst (iter n (Z * Z)
           (fun p : Z * Z =>
              match p with
                (r, i) => (r*(i+1), i+1)
              end) (1, 0))
.

Lemma ZfactS :
  forall n : Z,
    n >= 0 -> Zfact (n+1) = Zfact n * (n+1)
.
Proof.
intros n H.
destruct n.
 trivial.
 unfold Zfact in |- *.
   change (Zpos p + 1) at 1 with (Zpos (p + 1)) in |- *.
   unfold iter in |- *.
   replace (p + 1)%positive with (1 + p)%positive.
  rewrite
   (iter_pos_plus 1 p (Z * Z)
      (fun p0 : Z * Z => let (r, i) := p0 in (r * (i + 1), i + 1))
      (1, 0)).
    simpl in |- *.
    set
     (f := fun p0 : Z * Z => let (r, i) := p0 in (r * (i + 1), i + 1))
     in |- *.
    change (Zpos (p + 1)) with (Zpos p + 1) in |- *.
    rewrite Zpos_eq_Z_of_nat_o_nat_of_P with (p := p).
    apply
     iter_pos_better_invariant
      with
        (Inv := fun (n : nat) (a : Z * Z) =>
                fst (let (r, i) := a in (r * (i + 1), i + 1)) =
                fst a * (Z_of_nat n + 1)).
   intros n a.
     destruct a.
     unfold f in |- *.
     intro H'.
     simpl in |- *.
     simpl in H'.
     rewrite H'.
     rewrite Zmult_comm with (n := z).
     rewrite <- Zmult_assoc.
     rewrite <- Zmult_assoc.
     assert (z * (z0 + 1 + 1) = z * Zpos (P_of_succ_nat n + 1)).
    rewrite Zmult_plus_distr_r.
      rewrite H'.
      change (Zpos (P_of_succ_nat n + 1)) with (Z_of_nat (S n) + 1) in |- *.
      rewrite <- Zmult_plus_distr_r.
      assert (Z_of_nat n + 1 + 1 = Z_of_nat (S n) + 1).
     rewrite inj_S.
       trivial.
     rewrite H0.
       reflexivity.
    rewrite H0.
      reflexivity.
   trivial.
  apply Pplus_comm.
 absurd (Zneg p >= 0).
  intro.
    apply H.
    simpl in |- *.
    reflexivity.
  assumption.
Qed.

Lemma Zne_lt_gt :
forall n m : Z,
   n < m \/ n > m -> n <> m
.

Proof.
intros.
destruct H.
 apply Zlt_not_eq.
   assumption.
 unfold not in |- *.
   intros H0.
   rewrite H0 in H.
   apply (Zgt_irrefl _ H).
Qed.

Lemma Zne_is_ne_bool :
forall n m:Z,
   n <> m -> Zneq_bool n m = true
.

Proof.
intros.
assert (n < m \/ m < n).
 apply not_Zeq.
   assumption.
 unfold Zneq_bool in |- *.
   destruct H0.
  unfold Zlt in H0.
    induction (n ?= m).
   discriminate H0.
   trivial.
   discriminate H0.
  assert (n > m).
   apply Zlt_gt.
     assumption.
   unfold Zgt in H1.
     induction (n ?= m).
    discriminate.
    trivial.
    trivial.
Qed.

Lemma Zne_bool_is_ne :
forall n m:Z,
   Zneq_bool n m = true -> n <> m.

Proof.
intros.
apply Zne_lt_gt.
unfold Zneq_bool in H.
unfold Zlt, Zgt in |- *.
induction (n ?= m).
 discriminate H.
 left; trivial.
 right; trivial.
Qed.

Lemma Zmod_2_succ :
forall m:Z,
   m > 0 -> m mod 2 <> 0 ->
      (Zsucc m) mod 2 = 0.

Proof.
intros.
assert (0 <= m mod 2 < 2).
 apply Z_mod_lt.
   auto with zarith.
 destruct H1.
   assert (0 < m mod 2).
  assert (0 < m mod 2 \/ 0 = m mod 2).
   apply Zle_lt_or_eq.
     assumption.
   clear H1.
     destruct H3.
    assumption.
    absurd (0 = m mod 2).
     apply sym_not_eq.
       assumption.
     assumption.
  clear H0.
    assert (m mod 2 = 1).
   apply Zle_antisym.
    apply Zlt_succ_le.
      assumption.
    change 1 with (Zsucc 0) in |- *.
      apply Zlt_le_succ.
      assumption.
   clear H2 H3 H1.
     replace m with (2 * (m / 2) + m mod 2).
    rewrite H0.
      replace (Zsucc (2 * (m / 2) + 1)) with ((1 + m / 2) * 2).
     change ((1 + m / 2) * 2) with (0 + (1 + m / 2) * 2) in |- *.
       change 0 at 2 with (0 mod 2) in |- *.
       apply Z_mod_plus.
       auto with zarith.
     auto with zarith.
    symmetry  in |- *.
      apply Z_div_mod_eq.
      auto with zarith.
Qed.

Definition Zprime : Z -> Prop :=
   fun n: Z =>
      forall m : Z, 2 <= m < n -> n mod m <> 0
.

Lemma Zprime_sqrt_mod2 :
   forall n:Z,
      n mod 2 <> 0 ->
      (forall m: Z, m mod 2 <> 0 -> 3 <= m -> m * m <= n -> n mod m <> 0)
         -> Zprime n.

Proof.
Admitted.

End Mathematics.
